// Copyright 2018 CoreOS, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package util

func IntToPtr(x int) *int {
	return &x
}

func StrToPtr(s string) *string {
	return &s
}

func BoolToPtr(b bool) *bool {
	return &b
}

func NilOrEmpty(s *string) bool {
	return s == nil || *s == ""
}

func NotEmpty(s *string) bool {
	return s != nil && *s != ""
}

func IsTrue(b *bool) bool {
	return b != nil && *b
}

func IsFalse(b *bool) bool {
	return b != nil && !*b
}
