// Copyright 2020 Red Hat, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package types

import (
	"github.com/coreos/ignition/v2/config/shared/errors"
)

func validateMode(m *int) error {
	if m != nil && (*m < 0 || *m > 07777) {
		return errors.ErrFileIllegalMode
	}
	return nil
}

func validateModeSpecialBits(m *int) error {
	if m != nil {
		mode := uint32(*m)
		if mode&07000 != 0 {
			return errors.ErrModeSpecialBits
		}
	}
	return nil
}
