// Copyright 2019 Red Hat, Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package util

import (
	"reflect"
)

type MergesKeys interface {
	MergedKeys() map[string]string
}

type IgnoresDups interface {
	IgnoreDuplicates() map[string]struct{}
}

type Keyed interface {
	Key() string
}

// CallKey is a helper to call the Key() function since this needs to happen a lot
func CallKey(v reflect.Value) string {
	if v.Kind() == reflect.String {
		return v.Convert(reflect.TypeOf("")).Interface().(string)
	}
	return v.Interface().(Keyed).Key()
}
